--- ospfd/kroute.c.orig	2009-01-16 21:50:13.000000000 +0100
+++ ospfd/kroute.c	2009-04-22 15:39:54.000000000 +0200
@@ -1028,9 +1028,11 @@
 	struct sockaddr_in	prefix;
 	struct sockaddr_in	nexthop;
 	struct sockaddr_in	mask;
+#if !defined(__FreeBSD__)
 	struct sockaddr_rtlabel	sa_rl;
-	int			iovcnt = 0;
 	const char		*label;
+#endif /* !defined(__FreeBSD__) */
+	int			iovcnt = 0;
 
 	if (kr_state.fib_sync == 0)
 		return (0);
@@ -1039,11 +1041,15 @@
 	bzero(&hdr, sizeof(hdr));
 	hdr.rtm_version = RTM_VERSION;
 	hdr.rtm_type = action;
+#if !defined(__FreeBSD__)
 	hdr.rtm_priority = RTP_OSPF;
+#endif /* !defined(__FreeBSD__) */
 	if (action == RTM_CHANGE)
 		hdr.rtm_fmask = RTF_REJECT|RTF_BLACKHOLE;
+#if !defined(__FreeBSD__)
 	else
 		hdr.rtm_flags = RTF_MPATH;
+#endif /* !defined(__FreeBSD__) */
 	hdr.rtm_seq = kr_state.rtseq++;	/* overflow doesn't matter */
 	hdr.rtm_msglen = sizeof(hdr);
 	/* adjust iovec */
@@ -1086,6 +1092,7 @@
 	iov[iovcnt].iov_base = &mask;
 	iov[iovcnt++].iov_len = sizeof(mask);
 
+#if !defined(__FreeBSD__) /* FreeBSD has no route labeling. */
 	if (kroute->rtlabel != 0) {
 		sa_rl.sr_len = sizeof(sa_rl);
 		sa_rl.sr_family = AF_UNSPEC;
@@ -1102,6 +1109,7 @@
 		iov[iovcnt].iov_base = &sa_rl;
 		iov[iovcnt++].iov_len = sizeof(sa_rl);
 	}
+#endif /* !defined(__FreeBSD__) */
 
 
 retry:
@@ -1144,7 +1152,9 @@
 	struct rt_msghdr	*rtm;
 	struct sockaddr		*sa, *rti_info[RTAX_MAX];
 	struct sockaddr_in	*sa_in;
+#if !defined(__FreeBSD__)
 	struct sockaddr_rtlabel	*label;
+#endif
 	struct kroute_node	*kr;
 
 	mib[0] = CTL_NET;
@@ -1153,9 +1163,13 @@
 	mib[3] = AF_INET;
 	mib[4] = NET_RT_DUMP;
 	mib[5] = 0;
+#if !defined(__FreeBSD__) /* FreeBSD has no multiple routing tables */
 	mib[6] = 0;	/* rtableid */
 
 	if (sysctl(mib, 7, NULL, &len, NULL, 0) == -1) {
+#else
+	if (sysctl(mib, 6, NULL, &len, NULL, 0) == -1) {
+#endif /* !defined(__FreeBSD__) */
 		log_warn("sysctl");
 		return (-1);
 	}
@@ -1163,7 +1177,11 @@
 		log_warn("fetchtable");
 		return (-1);
 	}
+#if !defined(__FreeBSD__) /* FreeBSD has no multiple routing tables */
 	if (sysctl(mib, 7, buf, &len, NULL, 0) == -1) {
+#else
+	if (sysctl(mib, 6, buf, &len, NULL, 0) == -1) {
+#endif /* !defined(__FreeBSD__) */
 		log_warn("sysctl");
 		free(buf);
 		return (-1);
@@ -1233,6 +1251,7 @@
 			send_rtmsg(kr_state.fd, RTM_DELETE, &kr->r);
 			free(kr);
 		} else {
+#if !defined(__FreeBSD__) /* FreeBSD has no route labeling. */
 			if ((label = (struct sockaddr_rtlabel *)
 			    rti_info[RTAX_LABEL]) != NULL) {
 				kr->r.rtlabel =
@@ -1240,6 +1259,7 @@
 				kr->r.ext_tag =
 				    rtlabel_id2tag(kr->r.rtlabel);
 			}
+#endif /* !defined(__FreeBSD__) */
 			kroute_insert(kr);
 		}
 
@@ -1288,7 +1308,7 @@
 		switch (rtm->rtm_type) {
 		case RTM_IFINFO:
 			memcpy(&ifm, next, sizeof(ifm));
-			sa = (struct sockaddr *)(next + rtm->rtm_hdrlen);
+			sa = (struct sockaddr *)(next + sizeof(ifm));
 			get_rtaddrs(ifm.ifm_addrs, sa, rti_info);
 			if_change(ifm.ifm_index, ifm.ifm_flags, &ifm.ifm_data,
 			    (struct sockaddr_dl *)rti_info[RTAX_IFP]);
@@ -1323,7 +1343,9 @@
 	struct ifa_msghdr	*ifam;
 	struct sockaddr		*sa, *rti_info[RTAX_MAX];
 	struct sockaddr_in	*sa_in;
+#if !defined(__FreeBSD__)
 	struct sockaddr_rtlabel	*label;
+#endif /* !defined(__FreeBSD__) */
 	struct kroute_node	*kr, *okr;
 	struct in_addr		 prefix, nexthop;
 	u_int8_t		 prefixlen, prio;
@@ -1358,8 +1380,10 @@
 			sa = (struct sockaddr *)(rtm + 1);
 			get_rtaddrs(rtm->rtm_addrs, sa, rti_info);
 
+#if !defined(__FreeBSD__) /* XXX: FreeBSD has no multiple routing tables */
 			if (rtm->rtm_tableid != 0)
 				continue;
+#endif
 
 			if (rtm->rtm_pid == kr_state.pid) /* caused by us */
 				continue;
@@ -1447,6 +1471,7 @@
 				rtlabel_unref(kr->r.rtlabel);
 				kr->r.rtlabel = 0;
 				kr->r.ext_tag = 0;
+#if !defined(__FreeBSD__) /* FreeBSD has no route labeling. */
 				if ((label = (struct sockaddr_rtlabel *)
 				    rti_info[RTAX_LABEL]) != NULL) {
 					kr->r.rtlabel =
@@ -1454,6 +1479,7 @@
 					kr->r.ext_tag =
 					    rtlabel_id2tag(kr->r.rtlabel);
 				}
+#endif /* !defined(__FreeBSD__) */
 
 				if (kif_validate(kr->r.ifindex))
 					kr->r.flags &= ~F_DOWN;
@@ -1476,6 +1502,7 @@
 				kr->r.ifindex = ifindex;
 				kr->r.priority = prio;
 
+#if !defined(__FreeBSD__) /* FreeBSD has no route labeling */
 				if ((label = (struct sockaddr_rtlabel *)
 				    rti_info[RTAX_LABEL]) != NULL) {
 					kr->r.rtlabel =
@@ -1483,6 +1510,7 @@
 					kr->r.ext_tag =
 					    rtlabel_id2tag(kr->r.rtlabel);
 				}
+#endif /* !defined(__FreeBSD__) */
 
 				kroute_insert(kr);
 			}
@@ -1506,7 +1534,7 @@
 			break;
 		case RTM_IFINFO:
 			memcpy(&ifm, next, sizeof(ifm));
-			sa = (struct sockaddr *)(next + rtm->rtm_hdrlen);
+			sa = (struct sockaddr *)(next + sizeof(ifm));
 			get_rtaddrs(ifm.ifm_addrs, sa, rti_info);
 			if_change(ifm.ifm_index, ifm.ifm_flags, &ifm.ifm_data,
 			    (struct sockaddr_dl *)rti_info[RTAX_IFP]);
