--- ospfd/kroute.c.orig	2009-04-22 15:41:17.000000000 +0200
+++ ospfd/kroute.c	2009-04-22 15:44:37.000000000 +0200
@@ -68,7 +68,7 @@
 int	kr_change_fib(struct kroute_node *, struct kroute *, int, int);
 int	kr_delete_fib(struct kroute_node *);
 
-struct kroute_node	*kroute_find(in_addr_t, u_int8_t, u_int8_t);
+struct kroute_node	*kroute_find(in_addr_t, u_int8_t);
 struct kroute_node	*kroute_matchgw(struct kroute_node *, struct in_addr);
 int			 kroute_insert(struct kroute_node *);
 int			 kroute_remove(struct kroute_node *);
@@ -237,7 +237,6 @@
 		kn->r.prefixlen = kroute[i].prefixlen;
 		kn->r.nexthop.s_addr = kroute[i].nexthop.s_addr;
 		kn->r.flags = kroute[i].flags | F_OSPFD_INSERTED;
-		kn->r.priority = RTP_OSPF;
 		kn->r.ext_tag = kroute[i].ext_tag;
 		rtlabel_unref(kn->r.rtlabel);	/* for RTM_CHANGE */
 		kn->r.rtlabel = kroute[i].rtlabel;
@@ -261,10 +260,32 @@
 
 	kroute->rtlabel = rtlabel_tag2id(kroute->ext_tag);
 
-	kr = kroute_find(kroute->prefix.s_addr, kroute->prefixlen, RTP_OSPF);
-	if (kr != NULL && kr->next == NULL && krcount == 1)
-		/* single path OSPF route */
-		action = RTM_CHANGE;
+	kr = kroute_find(kroute->prefix.s_addr, kroute->prefixlen);
+
+	if (kr != NULL) {
+		if (kr->r.flags & F_KERNEL) {
+			/* a non-ospf route already exists. not a problem */
+			if (!(kr->r.flags & F_BGPD_INSERTED)) {
+				do {
+					kr->r.flags |= F_OSPFD_INSERTED;
+					kr = kr->next;
+				} while (kr);
+				return (0);
+			}
+			/*
+			 * XXX as long as there is no multipath support in
+			 * bgpd this is safe else we end up in a bad situation.
+			 */
+			/*
+			 * ospf route has higher pref
+			 * - reset flags to the ospf ones
+			 * - use RTM_CHANGE
+			 * - zero out ifindex (this is no longer relevant)
+			 */
+			action = RTM_CHANGE;
+		} else if (kr->next == NULL)	/* single path OSPF route */
+			action = RTM_CHANGE;
+	}
 
 	return (kr_change_fib(kr, kroute, krcount, action));
 }
@@ -272,9 +293,14 @@
 int
 kr_delete_fib(struct kroute_node *kr)
 {
-	if (kr->r.priority != RTP_OSPF)
-		log_warn("kr_delete_fib: %s/%d has wrong priority %d",
-		    inet_ntoa(kr->r.prefix), kr->r.prefixlen, kr->r.priority);
+	if (!(kr->r.flags & F_OSPFD_INSERTED))
+		return (0);
+
+	if (kr->r.flags & F_KERNEL) {
+		/* remove F_OSPFD_INSERTED flag, route still exists in kernel */
+		kr->r.flags &= ~F_OSPFD_INSERTED;
+		return (0);
+	}
 
 	if (send_rtmsg(kr_state.fd, RTM_DELETE, &kr->r) == -1)
 		return (-1);
@@ -290,8 +316,8 @@
 {
 	struct kroute_node	*kr, *nkr;
 
-	if ((kr = kroute_find(kroute->prefix.s_addr, kroute->prefixlen,
-	    RTP_OSPF)) == NULL)
+	if ((kr = kroute_find(kroute->prefix.s_addr, kroute->prefixlen)) ==
+	    NULL)
 		return (0);
 
 	while (kr != NULL) {
@@ -323,9 +349,10 @@
 	kr_state.fib_sync = 1;
 
 	RB_FOREACH(kr, kroute_tree, &krt)
-		if (kr->r.priority == RTP_OSPF)
-			for (kn = kr; kn != NULL; kn = kn->next)
+		if (!(kr->r.flags & F_KERNEL))
+			for (kn = kr; kn != NULL; kn = kn->next) {
 				send_rtmsg(kr_state.fd, RTM_ADD, &kn->r);
+			}
 
 	log_info("kernel routing table coupled");
 }
@@ -339,10 +366,12 @@
 	if (kr_state.fib_sync == 0)	/* already decoupled */
 		return;
 
-	RB_FOREACH(kr, kroute_tree, &krt)
-		if (kr->r.priority == RTP_OSPF)
-			for (kn = kr; kn != NULL; kn = kn->next)
+	RB_FOREACH(kr, kroute_tree, &krt) {
+		if (!(kr->r.flags & F_KERNEL))
+			for (kn = kr; kn != NULL; kn = kn->next) {
 				send_rtmsg(kr_state.fd, RTM_DELETE, &kn->r);
+			}
+	}
 
 	kr_state.fib_sync = 0;
 
@@ -353,7 +382,6 @@
 void
 kr_dispatch_msg(int fd, short event, void *bula)
 {
-	/* XXX this is stupid */
 	dispatch_rtmsg();
 }
 
@@ -504,10 +532,6 @@
 	struct rroute		 rr;
 	int			 redistribute = 0;
 
-	/* only the highest prio route can be redistributed */
-	if (kroute_find(kh->r.prefix.s_addr, kh->r.prefixlen, RTP_ANY) != kh)
-		return;
-
 	bzero(&rr, sizeof(rr));
 	rr.metric = UINT_MAX;
 	for (kn = kh; kn; kn = kn->next)
@@ -567,14 +591,6 @@
 		return (-1);
 	if (a->r.prefixlen > b->r.prefixlen)
 		return (1);
-
-	/* if the priority is RTP_ANY finish on the first address hit */
-	if (a->r.priority == RTP_ANY || b->r.priority == RTP_ANY)
-		return (0);
-	if (a->r.priority < b->r.priority)
-		return (-1);
-	if (a->r.priority > b->r.priority)
-		return (1);
 	return (0);
 }
 
@@ -586,27 +602,14 @@
 
 /* tree management */
 struct kroute_node *
-kroute_find(in_addr_t prefix, u_int8_t prefixlen, u_int8_t prio)
+kroute_find(in_addr_t prefix, u_int8_t prefixlen)
 {
 	struct kroute_node	s;
-	struct kroute_node	*kn, *tmp;
 
 	s.r.prefix.s_addr = prefix;
 	s.r.prefixlen = prefixlen;
-	s.r.priority = prio;
 
-	kn = RB_FIND(kroute_tree, &krt, &s);
-	if (kn && prio == RTP_ANY) {
-		tmp = RB_PREV(kroute_tree, &krt, kn);
-		while (tmp) {
-			if (kroute_compare(&s, tmp) == 0)
-				kn = tmp;
-			else 
-				break;
-			tmp = RB_PREV(kroute_tree, &krt, kn);
-		}
-	}
-	return (kn);
+	return (RB_FIND(kroute_tree, &krt, &s));
 }
 
 struct kroute_node *
@@ -636,6 +639,7 @@
 		 * Multipath route, add at end of list and clone the
 		 * ospfd inserted flag.
 		 */
+		kr->r.flags |= krm->r.flags & F_OSPFD_INSERTED;
 		while (krm->next != NULL)
 			krm = krm->next;
 		krm->next = kr;
@@ -845,12 +849,11 @@
 
 	/* we will never match the default route */
 	for (i = 32; i > 0; i--)
-		if ((kr = kroute_find(key & prefixlen2mask(i), i,
-		    RTP_ANY)) != NULL)
+		if ((kr = kroute_find(key & prefixlen2mask(i), i)) != NULL)
 			return (kr);
 
 	/* if we don't have a match yet, try to find a default route */
-	if ((kr = kroute_find(0, 0, RTP_ANY)) != NULL)
+	if ((kr = kroute_find(0, 0)) != NULL)
 			return (kr);
 
 	return (NULL);
@@ -1041,11 +1044,12 @@
 	bzero(&hdr, sizeof(hdr));
 	hdr.rtm_version = RTM_VERSION;
 	hdr.rtm_type = action;
+	hdr.rtm_flags = RTF_PROTO2;
 #if !defined(__FreeBSD__)
 	hdr.rtm_priority = RTP_OSPF;
 #endif /* !defined(__FreeBSD__) */
 	if (action == RTM_CHANGE)
-		hdr.rtm_fmask = RTF_REJECT|RTF_BLACKHOLE;
+		hdr.rtm_fmask = RTF_PROTO2|RTF_PROTO1|RTF_REJECT|RTF_BLACKHOLE;
 #if !defined(__FreeBSD__)
 	else
 		hdr.rtm_flags = RTF_MPATH;
@@ -1208,7 +1212,6 @@
 		}
 
 		kr->r.flags = F_KERNEL;
-		kr->r.priority = rtm->rtm_priority;
 
 		switch (sa->sa_family) {
 		case AF_INET:
@@ -1219,6 +1222,8 @@
 				kr->r.flags |= F_STATIC;
 			if (rtm->rtm_flags & RTF_DYNAMIC)
 				kr->r.flags |= F_DYNAMIC;
+			if (rtm->rtm_flags & RTF_PROTO1)
+				kr->r.flags |= F_BGPD_INSERTED;
 			if (sa_in != NULL) {
 				if (sa_in->sin_len == 0)
 					break;
@@ -1247,7 +1252,7 @@
 				break;
 			}
 
-		if (rtm->rtm_priority == RTP_OSPF)  {
+		if (rtm->rtm_flags & RTF_PROTO2)  {
 			send_rtmsg(kr_state.fd, RTM_DELETE, &kr->r);
 			free(kr);
 		} else {
@@ -1348,7 +1353,7 @@
 #endif /* !defined(__FreeBSD__) */
 	struct kroute_node	*kr, *okr;
 	struct in_addr		 prefix, nexthop;
-	u_int8_t		 prefixlen, prio;
+	u_int8_t		 prefixlen;
 	int			 flags, mpath;
 	u_short			 ifindex = 0;
 
@@ -1373,7 +1378,6 @@
 		flags = F_KERNEL;
 		nexthop.s_addr = 0;
 		mpath = 0;
-		prio = 0;
 
 		if (rtm->rtm_type == RTM_ADD || rtm->rtm_type == RTM_CHANGE ||
 		    rtm->rtm_type == RTM_DELETE) {
@@ -1398,8 +1402,6 @@
 			if (rtm->rtm_flags & RTF_MPATH)
 				mpath = 1;
 #endif
-			prio = rtm->rtm_priority;
-
 			switch (sa->sa_family) {
 			case AF_INET:
 				prefix.s_addr =
@@ -1419,6 +1421,8 @@
 					flags |= F_STATIC;
 				if (rtm->rtm_flags & RTF_DYNAMIC)
 					flags |= F_DYNAMIC;
+				if (rtm->rtm_flags & RTF_PROTO1)
+					flags |= F_BGPD_INSERTED;
 				break;
 			default:
 				continue;
@@ -1447,8 +1451,8 @@
 				continue;
 			}
 
-			if ((okr = kroute_find(prefix.s_addr, prefixlen, prio))
-			    != NULL) {
+			if ((okr = kroute_find(prefix.s_addr, prefixlen)) !=
+			    NULL) {
 				/* just add new multipath routes */
 				if (mpath && rtm->rtm_type == RTM_ADD)
 					goto add;
@@ -1462,6 +1466,13 @@
 					goto add;
 				}
 
+				/*
+				 * ospf route overridden by kernel. Preference
+				 * of the route is not checked because this is
+				 * forced -- most probably by a user.
+				 */
+				if (kr->r.flags & F_OSPFD_INSERTED)
+					flags |= F_OSPFD_INSERTED;
 				if (kr->r.flags & F_REDISTRIBUTED)
 					flags |= F_REDISTRIBUTED;
 				kr->r.nexthop.s_addr = nexthop.s_addr;
@@ -1500,7 +1511,6 @@
 				kr->r.nexthop.s_addr = nexthop.s_addr;
 				kr->r.flags = flags;
 				kr->r.ifindex = ifindex;
-				kr->r.priority = prio;
 
 #if !defined(__FreeBSD__) /* FreeBSD has no route labeling */
 				if ((label = (struct sockaddr_rtlabel *)
@@ -1516,8 +1526,8 @@
 			}
 			break;
 		case RTM_DELETE:
-			if ((kr = kroute_find(prefix.s_addr, prefixlen, prio))
-			    == NULL)
+			if ((kr = kroute_find(prefix.s_addr, prefixlen)) ==
+			    NULL)
 				continue;
 			if (!(kr->r.flags & F_KERNEL))
 				continue;
@@ -1529,6 +1539,14 @@
 				    " not found");
 				return (-1);
 			}
+			/*
+			 * last route is getting removed request the
+			 * ospf route from the RDE to insert instead
+			 */
+			if (okr == kr && kr->next == NULL &&
+			    kr->r.flags & F_OSPFD_INSERTED)
+				main_imsg_compose_rde(IMSG_KROUTE_GET, 0,
+				    &kr->r, sizeof(struct kroute));
 			if (kroute_remove(kr) == -1)
 				return (-1);
 			break;
